;;; linked-buffer.el --- One buffer as a view of another -*- lexical-binding: t -*-

;; This file is not part of Emacs

;; Author: Phillip Lord <phillip.lord@newcastle.ac.uk>
;; Maintainer: Phillip Lord <phillip.lord@newcastle.ac.uk>
;; Version: 0.5
;; Package-Requires: ((emacs "24")(m-buffer "0.5")(dash "2.5.0"))

;; The contents of this file are subject to the GPL License, Version 3.0.
;;
;; Copyright (C) 2014, Phillip Lord, Newcastle University
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;;
;; Sometimes, it would be nice to edit a file in two ways at once. For
;; instance, you might have a source file in a computational language with
;; richly marked documentation. As Emacs is a modal editor, you can edit one
;; in a mode for the computational language or for the marked up
;; documentation.
;;
;; One solution to this is to use one of the multiple-mode tools which are
;; available. The problem with this is that generally need some support from
;; the modes in question. In addition, they also require tools that work with
;; the mixed content; for example, Haskells literate mode.
;;
;; Linked buffers provide an alternative solution. Two linked buffers, by
;; default, the two share content but are otherwise independent. Therefore,
;; you can have two buffers open, each showing the content in different modes;
;; to switch modes, you simply switch buffers. The content, location of point,
;; and view are shared.
;;
;; However, linked-buffers also a bi-directional transformation between the
;; two. If this is done, then the two can have different but related text. It
;; is possible to configure the transformation for any two pairs of modes.
;;
;; Main entry points are `linked-buffer-split-window-right' and
;; `linked-buffer-split-window-below' both of which create a linked buffer in
;; the new window. Programmatically `linked-buffer-create', well, creates a
;; linked-buffer.
;;
;;; Configuration:
;;
;; linked-buffers are configurable in a large number of ways. It is possible
;; to control the nature of the transformation, the default buffer name that a
;; linked-buffer takes, and the file location (or not) of the linked-buffer.
;; For this release of linked-buffer currently, each buffer can only be linked
;; to a single buffer, although this restriction will be removed in later
;; versions.
;;
;; Configuration of a buffer happens in one of two places. First,
;; `linked-buffer-init' is run when a linked-buffer is first created. This
;; function should set the actual configuration `linked-buffer-config', and is
;; mostly designed for use as a file-local variable. All subsequent
;; configuration happens through `linked-buffer-config' which is an EIEIO
;; object and associated methods.
;;
;; Currently, I have only two concrete and one abstract configurations -- one
;; which copies all text exactly, but does not transfer text-properties (which
;; indirect-buffers do). There is a block-comment configuration which is
;; designed for syntaxes where beginning on line comments are required in
;; blocks of one buffer but not in the other. Finally, there is a concrete
;; extension of the block-comment configuration which is allows transformation
;; between Clojure code and latex.
;;
;; More configurations will be forth-coming -- next on the list will be
;; Emacs-Lisp to either asciidoc or markdown, so that this source can be made
;; literate.
;;
;; I think that the current configuration scheme is good enough for the
;; future, but only time will tell and it should still be considered
;; preliminary.
;;
;;; Status:
;;
;; This is an early release partly because I am interested in comments.
;; Hopefully, it will crash rather than hang Emacs. It currently performs
;; badly on large buffers, especially changes which make many small changes.
;; There are some outstanding bugs.

;;; Code:

(require 'eieio)

(defvar linked-buffer-init 'linked-buffer-default-init
  "Function that initializes a linked-buffer.
This should set up `linked-buffer-config' appropriately.")

;; In future versions, this may get turned into a list so that we can have
;; multiple linked buffers, but it is not clear how the user interface
;; functions such as `linked-buffer-swap-window' would work now.
(defvar linked-buffer-config nil
  "Configuration for linked-buffer.

This is object created by function linked-buffer-configuration',
which defines the way in which the text in the different buffers
is kept synchronized. This configuration is resiliant to changes
of mode in the current buffer.")

(make-variable-buffer-local 'linked-buffer-config)
(put 'linked-buffer-config 'permanent-local t)

(defvar linked-buffer-init-functions nil
  "A list of all functions that can be used as linked-buffer-init
  functions.")

(defun linked-buffer-config-name (buffer)
  "Given BUFFER, return a name for the configuration object."
  (format "linked: %s" buffer))

;;
;; Base Configuration:

;;
(defclass linked-buffer-configuration ()
  ((this-buffer
    :initarg :this-buffer)
   (that-buffer
    :initarg :that-buffer))
  "Configuration object for linked-buffer, which defines the mechanism
by which linking happens.")

(defgeneric linked-buffer-create (conf))
(defgeneric linked-buffer-convert (conf location))
(defgeneric linked-buffer-invert (conf that-buffer))

(defmethod linked-buffer-this ((conf linked-buffer-configuration))
  (oref conf :this-buffer))

(defmethod linked-buffer-that ((conf linked-buffer-configuration))
  (oref conf :that-buffer))

(defmethod linked-buffer-ensure-that ((conf linked-buffer-configuration))
  "Get the linked-buffer for this configuration
or create it if it does not exist."
  (or (linked-buffer-that conf)
      (linked-buffer-create conf)))

;;
;; Default Configuration:
;;
;; Two buffers with exactly the same contents, like an indirect buffer
;;
(defclass linked-buffer-default-configuration (linked-buffer-configuration)
  ((linked-file
    :initform nil
    :initarg :linked-file)
   (linked-mode
    :initform 'normal-mode
    :initarg :linked-mode))
  "Configuration which maintains two linked-buffers with the same contents.")

(defmethod linked-buffer-create ((conf linked-buffer-default-configuration))
  "Create the linked-buffer for this configuration.
Given a `linked-buffer-configuration' object, create the linked-buffer
appropriate for that configurationuration. It is the callers
responsibility to check that buffer has not already been
created."
  ;; make sure the world is ready for linked buffers
  (linked-buffer-ensure-hooks)
  ;; create linked-buffer
  (let* ((this-buffer
          (linked-buffer-this conf))
         (that-buffer
          (get-buffer-create
           (format "*linked: %s*"
                   (buffer-name
                    this-buffer))))
         (sec-mode (oref conf :linked-mode))
         (sec-file (oref conf :linked-file)))
    ;; make sure this-buffer knows about that-buffer
    (oset conf :that-buffer that-buffer)
    ;; insert the contents
    (linked-buffer-update-contents conf)
    ;; init that-buffer with mode, file and config
    ;; the mode must be init'd after adding content in case there are any
    ;; file-local variables need to be evaled
    (with-current-buffer that-buffer
      (when sec-mode
        (funcall sec-mode))
      (when sec-file
        (set-visited-file-name sec-file))
      (setq linked-buffer-config
            (linked-buffer-invert conf)))
    that-buffer))

(defmethod linked-buffer-invert ((conf linked-buffer-default-configuration))
  (linked-buffer-default-configuration
   (linked-buffer-config-name (linked-buffer-that conf))
   :this-buffer (oref conf :that-buffer)
   :that-buffer (oref conf :this-buffer)))

(defmethod linked-buffer-convert ((conf linked-buffer-default-configuration)
                                  location)
  "For this configuration, convert LOCATION to an equivalent location in
the linked-buffer."
  location)

(defmethod linked-buffer-clone ((conf linked-buffer-configuration))
  "Updates that-buffer to reflect the contents in this-buffer.

Currently, this is just a clone all method but may use regions in future."
  (with-current-buffer (oref conf :that-buffer)
    (erase-buffer)
    (insert
     (save-restriction
       (with-current-buffer (oref conf :this-buffer)
         (widen)
         (buffer-substring-no-properties
          (point-min)
          (point-max)))))))

(defun linked-buffer-default-init ()
  "Default init function.
See `linked-buffer-init' for details."
  (setq linked-buffer-config
        (linked-buffer-default-configuration
         (linked-buffer-config-name (current-buffer))
         :this-buffer (current-buffer))))

(add-to-list 'linked-buffer-init-functions
             'linked-buffer-default-init)

;;
;; End the configuration section.
;;


(defmacro linked-buffer-when-linked (&rest body)
  "Evaluate BODY when in a linked-buffer."
  `(when (and
          linked-buffer-config
          (linked-buffer-that
           linked-buffer-config)
          (buffer-live-p
           (linked-buffer-that
            linked-buffer-config)))
     ,@body))

(defun linked-buffer-ensure-hooks ()
  "Ensures that the hooks that this mode requires are in place."
  (add-hook 'post-command-hook
            'linked-buffer-post-command-hook)
  ;; after and before-change functions are hooks (with args) even if they are
  ;; not named as such.
  (add-hook 'after-change-functions
            'linked-buffer-after-change-function)
  (add-hook 'before-change-functions
            'linked-buffer-before-change-function)
  (add-hook 'after-save-hook
            'linked-buffer-after-save-hook))

(defvar linked-buffer-log t)
(defmacro linked-buffer-log (&rest rest)
  "Log REST."
  `(when linked-buffer-log
     (linked-buffer-when-linked
      (let ((msg
             (concat
              (format ,@rest)
              "\n")))
        (with-current-buffer
            (get-buffer-create "*linked-buffer-log*")
          (goto-char (point-max))
          (insert msg))))))

(defvar linked-buffer-emergency nil)

(defun linked-buffer-emergency ()
  "Stop linked-buffer from working due to code problem"
  (interactive)
  (setq linked-buffer-emergency t))

(defun linked-buffer-unemergency ()
  (interactive)
  (setq linked-buffer-emergency nil))

(defvar linked-buffer-saving-p nil)

(defun linked-buffer-after-save-hook ()
  (linked-buffer-when-linked
   ;; don't want to recurse!
   (when (not linked-buffer-saving-p)
     (let ((linked-buffer-saving-p t))
       (with-current-buffer
           (linked-buffer-that linked-buffer-config)
         (when (buffer-file-name)
           (save-buffer)))))))

(defun linked-buffer-post-command-hook ()
  "Update point according to config, with error handling."
  ;;(message "Entering post-command-hook")
  (unless linked-buffer-emergency
    (condition-case err
        (linked-buffer-post-command-hook-1)
      (error
       (linked-buffer-hook-fail err "post-command-hook")))))

(defun linked-buffer-post-command-hook-1 ()
  "Update point according to config."
  (progn
    (linked-buffer-when-linked
     (linked-buffer-update-point linked-buffer-config))))

(defun linked-buffer-hook-fail (err hook)
  "Give an informative message when we have to fail.
ERR is the error. HOOK is the hook type."
  (message "linked-buffer mode has failed on %s hook: %s "
           hook (error-message-string err))
  (linked-buffer-emergency)
  (with-output-to-temp-buffer "*linked-buffer-fail*"
    (princ "There has been an error in linked-buffer-mode.\n")
    (princ "The following is debugging information\n\n")
    (princ (error-message-string err)))
  (select-window (get-buffer-window "*linked-buffer-fail*")))

(defun linked-buffer-move-linked-window ()
  "Move the linked-buffer into the current window.
If the linked-buffer is currently being displayed in another
window, then the current-buffer will be moved into that window.
See also `linked-buffer-swap-buffer-windows'."
  (interactive)
  (let ((before-window-start
         (window-start (get-buffer-window)))
        (before-window-point
         (point)))
    (linked-buffer-swap-buffer-windows
     (current-buffer)
     (linked-buffer-that linked-buffer-config))
    (set-window-start
     (selected-window)
     before-window-start)
    (goto-char before-window-point)))

(defun linked-buffer-swap-linked-window ()
  "Swap the window of the buffer and linked-buffer.
If both are current displayed, swap the windows they
are displayed in, which keeping current buffer.
See also `linked-buffer-move-linked-window'."
  (interactive)
  (linked-buffer-swap-buffer-windows
   (current-buffer)
   (linked-buffer-that linked-buffer-config))
  (when (window-live-p
         (get-buffer-window
          (current-buffer)))
    (select-window
     (get-buffer-window
      (current-buffer)))))

(defun linked-buffer-swap-buffer-windows (a b)
  "Swaps the window that two buffers are displayed in.
A and B are the buffers."
  (let ((window-a (get-buffer-window a))
        (window-b (get-buffer-window b)))
    (when window-a
      (set-window-buffer
       window-a b))
    (when window-b
      (set-window-buffer
       window-b a))))

(defun linked-buffer-ensure-init ()
  "Ensure that the `linked-buffer-init' has been run."
  (unless (and linked-buffer-config
               (slot-boundp
                linked-buffer-config :that-buffer)
               (buffer-live-p (linked-buffer-that
                               linked-buffer-config)))
    (funcall linked-buffer-init)))

(defun linked-buffer-init-create ()
  "Create the linked-buffer for current-buffer."
  (linked-buffer-ensure-init)
  (linked-buffer-create linked-buffer-config))

(defun linked-buffer-create-in-selected-window ()
  "Create a linked buffer and move it to the current window."
  (interactive)
  (let ((before-window-start
         (window-start (get-buffer-window)))
        (before-window-point
         (point)))
    (linked-buffer-ensure-init)
    (set-window-buffer
     (selected-window)
     (linked-buffer-create linked-buffer-config))
    (set-window-start
     (selected-window)
     before-window-start)
    (goto-char before-window-point)))

(defun linked-buffer-split-window-below ()
  "Create a linked buffer in a new window below."
  (interactive)
  (linked-buffer-ensure-init)
  (set-window-buffer
   (split-window-below)
   (linked-buffer-create linked-buffer-config)))

(defun linked-buffer-split-window-right ()
  "Create a linked buffer in a new window right."
  (interactive)
  (linked-buffer-ensure-init)
  (set-window-buffer
   (split-window-right)
   (linked-buffer-create linked-buffer-config)))

(defun linked-buffer-after-change-function (&rest rest)
  "Run change update according to `linked-buffer-config'.
Errors are handled. REST is currently just ignored."
  (unless linked-buffer-emergency
    (condition-case err
        (linked-buffer-after-change-function-1 rest)
      (error
       (linked-buffer-hook-fail err "after change")))))

(defun linked-buffer-after-change-function-1 (rest)
  "Run change update according to `linked-buffer-config'.
REST is currently just ignored."
  (linked-buffer-when-linked
   (linked-buffer-log
    "Updating after-change (current:linked:rest): %s,%s,%s"
    (current-buffer)
    (linked-buffer-that linked-buffer-config) rest)
   (linked-buffer-update-contents linked-buffer-config)))

(defun linked-buffer-before-change-function (&rest rest)
  "Run before change update.
REST is currently ignored. Currently this does nothing."
  (unless linked-buffer-emergency
    (condition-case err
        (lambda ())
      (error
       (linked-buffer-hook-fail err "before change")))))

(defun linked-buffer-update-contents (conf)
  "Update the contents of that-buffer with the contents of this-buffer.
Update mechanism depends on CONF."
  (unwind-protect
      (progn
        (setq inhibit-read-only t)
        (linked-buffer-log
         "Update config: %s" linked-buffer-config)
        (linked-buffer-clone conf))
    (setq inhibit-read-only nil)))

(defun linked-buffer-update-point (conf)
  "Update the location of point in that-buffer to reflect this-buffer.
This also attempts to update any windows so that they show the
same top-left location. Update details depend on CONF."
  (let* ((from-point
          (linked-buffer-convert
           conf
           (with-current-buffer
               (linked-buffer-this conf)
             (point))))
         (from-window-start
          (linked-buffer-convert
           conf
           (window-start
            (get-buffer-window
             (linked-buffer-this conf))))))
    ;; clone point in buffer important when the buffer is NOT visible in a
    ;; window at all
    (with-current-buffer
        (linked-buffer-that conf)
      (goto-char from-point))
    ;; now clone point in all the windows that are showing the buffer
    ;; and set the start of the window which is a reasonable attempt to show
    ;; the same thing.
    (mapc
     (lambda (window)
       (with-selected-window window
         (progn
           (goto-char from-point)
           (set-window-start window from-window-start))))
     (get-buffer-window-list (linked-buffer-that conf)))))

;;
;; Minor mode
;;

(defvar linked-buffer-mode-map (make-sparse-keymap)
  "Keymap for linked-buffer-minor-mode")

(define-key linked-buffer-mode-map
  (kbd "C-c ,s") 'linked-buffer-swap-linked-window)

(define-key linked-buffer-mode-map
  (kbd "C-c ,h") 'linked-buffer-move-linked-window)

(define-minor-mode linked-buffer-mode
  :lighter "lb"
  :keymap linked-buffer-mode-map)

(easy-menu-change
 '("Edit")
 "Linked"
 '(["Create Here" linked-buffer-create-in-selected-window]
   ["Split Below" linked-buffer-split-window-below]
   ["Split Right" linked-buffer-split-window-right]
   ["Move Here" linked-buffer-move-linked-window :active linked-buffer-config]
   ["Swap" linked-buffer-swap-buffer-windows :active linked-buffer-config]))

(defun linked-buffer-insert-file-local (init-function)
  (interactive
   (list (completing-read
          "Linked-Buffer init function: "
          (mapcar
           'symbol-name
           linked-buffer-init-functions)
          'identity 'confirm)))
  (save-excursion
    (goto-char (point-max))
    (let ((start (point)))
      (insert
       (format
        "\nLocal Variables:\nlinked-buffer-init: %s\nEnd:\n" init-function))
      (comment-region start (point)))))

(defvar linked-buffer-start-mode-map (make-sparse-keymap))

(define-key linked-buffer-start-mode-map
  (kbd "C-c ,b") 'linked-buffer-split-window-below)

(define-key linked-buffer-start-mode-map
  (kbd "C-c ,r") 'linked-buffer-split-window-right)

(define-key linked-buffer-start-mode-map
  (kbd "C-c ,f") 'linked-buffer-insert-file-local)

(define-key linked-buffer-start-mode-map
  (kbd "C-c ,c") 'linked-buffer-create-in-selected-window)


(define-minor-mode linked-buffer-start-mode
  :lighter ""
  :keymap linked-buffer-start-mode-map)

(define-globalized-minor-mode global-linked-buffer-start-mode
  linked-buffer-start-mode
  linked-buffer-start-on)

(defun linked-buffer-start-on ()
  (linked-buffer-start-mode 1))

;;
;; Test functions!
;;
(defun linked-buffer-batch-clone-and-save-with-config (filename init)
  "Open FILENAME, set INIT function, then clone and save.

This function does potentially evil things if the file or the
linked-buffer is open already."
  (with-current-buffer
      (find-file-noselect filename)
    (setq linked-buffer-init init)
    (with-current-buffer
        (linked-buffer-init-create)
      (save-buffer)
      (kill-buffer))
    (kill-buffer)))

(defun linked-buffer-batch-clone-with-config
  (filename init)
  "Open FILENAME, set INIT function, then clone.

Return the linked-buffer contents without properties."
  (let ((retn nil))
    (with-current-buffer
        (find-file-noselect filename)
      (setq linked-buffer-init init)
      (with-current-buffer
          (linked-buffer-init-create)
        (setq retn
              (buffer-substring-no-properties
               (point-min)
               (point-max)))
        (set-buffer-modified-p nil)
        (kill-buffer))
      (set-buffer-modified-p nil)
      (kill-buffer))
    retn))

(defun linked-buffer-test-after-change-function ()
  "Run the change functions out of the command loop.
Using this function is the easiest way to test an new
`linked-buffer-clone' method, as doing so in the command loop is
painful for debugging. Set variable `linked-buffer-emergency' to
true to disable command loop functionality."
  (interactive)
  (linked-buffer-after-change-function-1 nil))

(defun linked-buffer-test-post-command-hook ()
  "Run the post-command functions out of the command loop.
Using this function is the easiest way to test an new
`linked-buffer-convert' method, as doing so in the command loop is
painful for debugging. Set variable `linked-buffer-emergency' to
true to disable command loop functionality."
  (interactive)
  (linked-buffer-post-command-hook-1))

(defun linked-buffer-test-reinit ()
  "Recall the init function regardless of current status.
This can help if you have change the config object and need
to make sure there is a new one."
  (interactive)
  (funcall linked-buffer-init))

(provide 'linked-buffer)

;;; linked-buffer.el ends here
